/* GNU POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef _POP3D_H
#define _POP3D_H	1
#include "config.h"

/* The version */
#define VERSION		"0.10.3"

/* The implementation */
#define	IMPL		"GNU POP3"

/* You can edit the messages the POP server prints out here */

/* Initial greeting */
#define WELCOME		"Welcome to " IMPL " " VERSION

/* A command that doesn't exist */
#define BAD_COMMAND	"Invalid command"

/* Incorrect number of arguments passed to a command */
#define BAD_ARGS	"Invalid arguments"

/* Command issued in wrong state */
#define BAD_STATE	"Incorrect state"

/* An action on a message that doesn't exist */
#define NO_MESG		"No such message"

/* A command that is known but not implemented */
#define NOT_IMPL	"Not implemented"

/* Invalid username or password */
#define BAD_LOGIN	"Bad login"

/* User authenticated, but mailbox is locked */
#define MBOX_LOCK	"Mailbox in use"

/* The command argument was > 40 characters */
#define TOO_LONG	"Argument too long"

/* User authenticated, but we can't place a lock on mailbox */
#define LOCK_FAILED	"Unable to lock mailbox"

/* Positioning in the mailbox file returned an error. File may be corrupt */
#define FILEPOS		"Error in mailbox file. File may be corrupted"

/* APOP password file, without .db or .passwd, which are added based on file
   type automatically */
#define APOP_PASSFILE	"/etc/apop"

/* Size of the MD5 digest for APOP */
#define APOP_DIGEST	70

/* Longest legal POP command */
#define POP_MAXCMDLEN	255

#ifdef USE_VIRTUAL

/* Define UID for using for manipulating mailboxes of virtual users.
   The mailboxes for virtual domains do not have corresponding Unix
   accounts; they do not have unique UIDs; so they could be owned
   by one VIRTUAL_UID. */
#define VIRTUAL_UID	8

/* VIRTUAL_MAILPATH is the path to the sub-directories containing
   the mailboxes for each virtual domain (or realm). It has a host,
   domain or realm name as a sub-directory which contain the mailbox
   files; for example: /var/spool/virtual/foo.bar/username where
   "username" is the mailbox name. (By default, this is
   /var/spool/virtual.) */

#define VIRTUAL_MAILPATH "/var/spool/virtual"

/* VIRTUAL_PASSWORDS_PATH is the first part (or prefix) of the path to
   the sub-directories containing the password files for each virtual
   domain. By default, this is /etc/virtual.) */

#define VIRTUAL_PASSWORDS_PATH  "/etc/virtual"

/* VIRTUAL_PASSWORD_FNAME is the file name (or suffix) for the virtual
   domain's password file. (By default, this is "passwd". It could be set
   to "mail/mailboxes", for example.) The full path to this password file
   is built by concatenating
    VIRTUAL_PASSWORDS_PATH "/" domainname "/" VIRTUAL_PASSWORD_FNAME
   where domainname is the name of the virtual host/realm that the
   virtual user is a part of. */

#define VIRTUAL_PASSWORD_FNAME  "passwd"

#endif /* USE_VIRTUAL */

#define _GNU_SOURCE
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <ctype.h>
#include "md5.h"
#include "md5_loc.h"

#ifdef USE_LOCAL_GETOPT_H
#include "getopt.h"
#else
#include <getopt.h>
#endif

#ifdef USE_IPV6
#define AF_IN AF_INET6
#else
#define AF_IN AF_INET
#endif

/* new flag for getaddrinfo */
#ifndef AI_CANONIDN
#define AI_CANONIDN 0
#endif

/* Maximum length of a hostname (is this defined somewhere else?) */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN        256
#endif

/* For Berkley DB2 APOP password file */
#ifdef BDB2
#include <db2/db.h>
#endif

/* The path to the mail spool files */
/* first pull in paths.h if available */
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif

/* then define _PATH_MAILDIR if not already defined */
#ifndef _PATH_MAILDIR
/* (WLS) This is a quick hack which really should be done by autoconf */
#ifdef __SVR4
#define _PATH_MAILDIR      "/var/spool/mail"
#else
#define _PATH_MAILDIR      "/usr/spool/mail"
#endif
#endif

/* currently PAM is only used for regular system accounts */
#ifdef USE_PAM
#include <security/pam_appl.h>
#endif

#ifdef HAVE_SHADOW_H
#include <shadow.h>
#endif

#define AUTHORIZATION	0
#define TRANSACTION	1
#define UPDATE		2

#define INTERACTIVE	0
#define DAEMON		1

#define OK		0
#define ERR_WRONG_STATE	1
#define ERR_BAD_ARGS	2
#define ERR_BAD_LOGIN	3
#define ERR_NO_MESG	4
#define ERR_NOT_IMPL	5
#define ERR_BAD_CMD	6
#define ERR_MBOX_LOCK	7
#define ERR_TOO_LONG	8
#define ERR_NO_MEM	9
#define ERR_DEAD_SOCK	10
#define ERR_SIGNAL	11
#define ERR_FILE        12
#define ERR_NO_OFILE    13
#define ERR_TIMEOUT	14
#define ERR_LOCK_FAILED 15
#define ERR_FILEPOS     16

#ifdef IP_BASED_VIRTUAL
#define ERR_IP_BASED	23
#endif /* IP_BASED_VIRTUAL */

#define ERR_MDN         30

#define TRUE 1
#define FALSE 0

typedef struct _pop3_message
{
  unsigned int size;
  fpos_t header;
  short deleted;
  unsigned char md5digest[16];
  fpos_t msgstart;				/* ptr to "From " */
  char status[8];				/* RFC 2076 Status: OR */
  short chng;					/* Header changed: has MDN been sent / unread msg been read/ new msg in mailbox */
}

message;

int debug;						/* 2 hex digits (child, daemon) as bit mask with lower bits for less logging */

unsigned int port;
unsigned int timeout;
int state;
int cursor;
int num_messages;
message *messages;
char *mailbox;
char *lockfile;
char *username;
uid_t uid;						/* uid can be set via command-line */
gid_t gid;						/* gid can be set via command-line */

struct sigaction signalaction;

#ifdef USE_VIRTUAL
char *virtualdomain;
#endif
FILE *mbox;
FILE *lock;
int ifile;
FILE *ofile;
time_t curr_time;
char *md5shared;

volatile sig_atomic_t child_procs;
int maxchildren;
volatile sig_atomic_t exit_flag;
volatile sig_atomic_t last_signal;	/* set by pop3_signal */
volatile pid_t *childpids;		/* table of started child processes */
char peeraddr[(INET6_ADDRSTRLEN + 1) * sizeof (char)];	/* IP address of current client */
volatile double time_used;		/* time used for commands in current client */
volatile int time_calls;		/* number of commands received in current client */
unsigned long int total_sessions;	/* no. of total connections */

pid_t daemonpid;				/* pid of main process, so children may use sigqueue to it */

char *useroptions;				/* from USER <name>--<options> */
char *tmppath;					/* for MDNs */
int sock;						/* for main waiting */

int mdn;						/* Message Disposition Notification 0 or 1 */
int mdncount;					/* no of MDNs sent this session */
int msgstatus;					/* Status: header to be updated 0 or 1 */
int ipv6enabled;				/* IPv6 enabled and used */

int pop3_dele (const char *arg);
int pop3_list (const char *arg);
int pop3_noop (const char *arg);
int pop3_quit (const char *arg);
int pop3_retr (const char *arg);
int pop3_rset (const char *arg);
int pop3_stat (const char *arg);
int pop3_top (const char *arg);
int pop3_uidl (const char *arg);
#ifdef USE_VIRTUAL
struct passwd *getvirtualpwnam (const char *username, const char *domainname);
char *pop3_splitname (char *arg, char *useroptions);
#endif
int pop3_user (const char *arg);
int pop3_apop (const char *arg);
int pop3_finishlogin (struct passwd *pw);
int pop3_auth (const char *arg);
int pop3_capa (const char *arg);
char *pop3_args (const char *cmd);
char *pop3_cmd (const char *cmd);
int pop3_mesg_exist (int mesg);
int pop3_abquit (int reason);
int pop3_lock (void);
int pop3_unlock (void);
int pop3_getsizes (void);
int pop3_mainloop (int infile, int outfile, char **argv);
int pop3_daemon (int maxchildren, char **argv);
void pop3_usage (char *argv0);
void pop3_signal (int sig, siginfo_t * siginfo, void *ucontext);
int pop3_childpidsupdate (int func, pid_t pid);
void pop3_daemon_init (void);
char *pop3_apopuser (const char *user);
char *pop3_readline (int fd);
void pop3_freepw (struct passwd *pwdentry);
int pop3_sendmdn (char *returnpath, char *envto, char *delivto, char *subject, char *date, char *msgid, char *sender);
char *pop3_pickupheader (const char *haystack, const char *needle);
int pop3_recover_oldmailbox (char *mailbox, FILE * mbox);
void pop3_syslog (int priority, const char *msg, const int flush);
void get_now (struct timeval *time, int debug);
double time_to_seconds (struct timeval *tstart, struct timeval *tfinish);
#endif /* _POP3D_H */
